#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include "types.h"

/* defines */

#define SCREEN_WIDTH    640
#define SCREEN_HEIGHT   400
#define TITLE           "BRESENHAM LINE DRAWING"

/* external variables */

/* None */

/* global variables */

/* None */

/* local variables */

static SDL_Surface    *pScr; /* the single SDL surface used */
static COLOUR_TYPE    Colour;

/* local prototypes */
static void drawdot(U32 x0, U32 y0);
static void octant0(U32 x0, U32 y0, U32 deltax, U32 deltay, S8 xdirection);
static void octant1(U32 x0, U32 y0, U32 deltax, U32 deltay, S8 xdirection);
static void drawline(S32 x0, S32 y0, S32 x1, S32 y1, COLOUR_TYPE colour);
static void vectorsup(S32 xcenter, S32 ycenter, S32 xlength, S32 ylength, COLOUR_TYPE colour);
static void redraw(void);
static void cls(void);

/* sample program to draw four rectangles full of lines */
/* using bresenham line drawing algo                    */
int main(int argc, char *argv[])
{
    S32          Width = SCREEN_WIDTH;
    S32          Height = SCREEN_HEIGHT;
    S32          ColorDepth = 32;
    U8           Done = 0;
    SDL_Event    Event = {0};
    COLOUR_TYPE  Setlinecolour;

    if(SDL_Init(SDL_INIT_EVERYTHING) < 0)
    {
        printf("Unable to init SDL: %s\n", SDL_GetError());
        SDL_Quit();
        return(1);
    }

    SDL_WM_SetCaption(TITLE, NULL);

    pScr = SDL_SetVideoMode(Width, Height, ColorDepth, SDL_HWSURFACE | SDL_HWPALETTE);

    if(pScr == NULL)
    {
        printf("Unable to set video: %s\n", SDL_GetError());
        SDL_Quit();
        return(1);
    }

    while(Done != 1)
    {
        /* draw each of the four rectangles full of vectors */
        Setlinecolour.r = 255;
        Setlinecolour.g = 0;
        Setlinecolour.b = 0;
        vectorsup((Width / 4), (Height / 4), (Width / 4) , (Height / 4) , Setlinecolour);

        Setlinecolour.r = 0;
        Setlinecolour.g = 255;
        Setlinecolour.b = 0;
        vectorsup((Width * 3 / 4), (Height / 4), (Width / 4) , (Height / 4) , Setlinecolour);

        Setlinecolour.r = 0;
        Setlinecolour.g = 0;
        Setlinecolour.b = 255;
        vectorsup((Width / 4), (Height * 3 / 4), (Width / 4) , (Height / 4) , Setlinecolour);

        Setlinecolour.r = 255;
        Setlinecolour.g = 0;
        Setlinecolour.b = 0;
        vectorsup((Width * 3 / 4), (Height * 3 / 4), (Width / 4) , (Height / 4) , Setlinecolour);

        /* update screen */
        redraw();
        /* clear screen buffer for next write */
        cls();

        SDL_PollEvent(&Event);

        /* is it time to quit ? */
        if(Event.type == SDL_QUIT)
        {
            SDL_Quit();
            Done = 1;
        }

        SDL_Delay(10); /* so it consumes less processing power */

    } /* while */
}

/****************************************************************/
/* Name: cls                                                    */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* Clears the screen to black                                   */
/*                                                              */
/****************************************************************/
static void cls(void)
{
  SDL_FillRect(pScr, NULL, 0);
}

/****************************************************************/
/* Name: redraw                                                 */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* Updates the screen.  Has to be called to view new pixels,    */
/* but use only after drawing the whole screen because it's     */
/* slow.                                                        */
/*                                                              */
/****************************************************************/
static void redraw(void)
{
  SDL_UpdateRect(pScr, 0, 0, 0, 0);
}

/****************************************************************/
/* Name: vectorsup                                              */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* subroutine to draw a rectangle full of vectors, of the       */
/* specified length and colour, around the specified rectangle  */
/* center.                                                      */
/*                                                              */
/****************************************************************/
static void vectorsup(S32 xcenter, S32 ycenter, S32 xlength, S32 ylength, COLOUR_TYPE colour)
{
    S32 workingx;
    S32 workingy;

    /* lines from the center to top of rectangle */
    workingx = xcenter - xlength;
    workingy = ycenter - ylength;
    for( ; workingx < (xcenter + xlength); workingx++)
    {
        drawline(xcenter, ycenter, workingx, workingy, colour);
    }

    /* lines from the center to right of rectangle */
    workingx = xcenter + xlength - 1;
    workingy = ycenter - ylength;
    for( ; workingy < (ycenter + ylength); workingy++)
    {
        drawline(xcenter, ycenter, workingx, workingy, colour);
    }

    /* lines from the center to bottom of rectangle */
    workingx = xcenter + xlength - 1;
    workingy = ycenter + ylength - 1;
    for( ; workingx >= (xcenter - xlength); workingx--)
    {
        drawline(xcenter, ycenter, workingx, workingy, colour);
    }

    /* lines from the center to left of rectangle */
    workingx = xcenter - xlength;
    workingy = ycenter + ylength - 1;
    for( ; workingy >= (ycenter - ylength); workingy--)
    {
        drawline(xcenter, ycenter, workingx, workingy, colour);
    }


}
/****************************************************************/
/* Name: drawline                                               */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* draw a line on the screen. x0, y0 coords are one end if the  */
/* line and x1, y1 coords are the other end of the line to draw */
/* and colour is the colour to draw the line                    */
/*                                                              */
/****************************************************************/
static void drawline(S32 x0, S32 y0, S32 x1, S32 y1, COLOUR_TYPE linecolour)
{
    S32 deltax;
    S32 deltay;
    S32 temp;

    /* set line colour*/
    Colour.r = linecolour.r;
    Colour.g = linecolour.g;
    Colour.b = linecolour.b;

    /*  save half the line-drawing cases by swapping y0 with y1
        and x0 with x1 if y0 is greater than y1. As a result, deltay
        is always > 0, and only the octant 0-3 cases need to be handled */

    if(y0 > y1)
    {
        temp = y0;
        y0 = y1;
        y1 = temp;

        temp = x0;
        x0 = x1;
        x1 = temp;
    }

    /* calculate line lengths */
    deltax = x1 - x0;
    deltay = y1 - y0;

    /* handle as four separate cases, for the four octants in which */
    /* y1 is greater then y0 */
    if(deltax > 0)
    {
        if(deltax > deltay)
        {
            octant0(x0,y0,deltax,deltay,1);
        }
        else
        {
            octant1(x0,y0,deltax,deltay,1);
        }
    }
    else
    {
        deltax = -deltax;       /* absolute value of deltax */
        if(deltax > deltay)
        {
            octant0(x0,y0,deltax,deltay,-1);
        }
        else
        {
            octant1(x0,y0,deltax,deltay,-1);
        }
    }
}

/****************************************************************/
/* Name: octant1                                                */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* draw a line in octant 1 or 2 ( |deltaX| < DeltaY )           */
/* start drawing line at x0, y0 for length of deltax , deltay   */
/* in direction of xdirection.                                  */
/*                                                              */
/****************************************************************/
static void octant1(U32 x0, U32 y0, U32 deltax, U32 deltay, S8 xdirection)
{
    S32 DeltaXx2;               /* holds delta between two midpoint line values */
    S32 DeltaXx2MinusDeltaYx2;  /* delta between two midpoint line values after overflow */
    S32 ErrorTerm;              /* error value between real line and drawn line */

    /* setup initial error term and values used inside drawing loop */
    DeltaXx2 = deltax * 2;
    DeltaXx2MinusDeltaYx2 = DeltaXx2 - (S32)(deltay *2);
    ErrorTerm = DeltaXx2 - (S32)deltay; /* delta between start line point and first midpoint value */

    /* draw the line */
    drawdot(x0, y0);        /* draw the first pixel */
    while(deltay--)
    {
        /* see if its time to advance the x coordinates */
        if(ErrorTerm > 0)
        {
            /* advance the x coordinate & adjust the error term back down */
            x0 += xdirection;
            ErrorTerm += DeltaXx2MinusDeltaYx2;
        }
        else
        {
            /* add to the error time */
            ErrorTerm += DeltaXx2;
        }

        y0++; /* advance the y coordinate */
        drawdot(x0, y0);
    }
};

/****************************************************************/
/* Name: octant0                                                */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* draw a line in octant 0 or 3 ( |deltaX| >= DeltaY )          */
/* start drawing line at x0, y0 for length of deltax , deltay   */
/* in direction of xdirection.                                  */
/*                                                              */
/****************************************************************/
static void octant0(U32 x0, U32 y0, U32 deltax, U32 deltay, S8 xdirection)
{
    S32 DeltaYx2;               /* holds delta between two midpoint line values */
    S32 DeltaYx2MinusDeltaXx2;  /* delta between two midpoint line values after overflow */
    S32 ErrorTerm;              /* error value between real line and drawn line */

    /* setup initial error term and values used inside drawing loop */
    DeltaYx2 = deltay * 2;
    DeltaYx2MinusDeltaXx2 = DeltaYx2 - (S32)(deltax *2);
    ErrorTerm = DeltaYx2 - (S32)deltax; /* delta between start line point and first midpoint value */

    /* draw the line */
    drawdot(x0, y0);        /* draw the first pixel */
    while(deltax--)
    {
        /* see if its time to advance the y coordinates */
        if(ErrorTerm > 0)
        {
            /* advance the y coordinate & adjust the error term back down */
            y0++;
            ErrorTerm += DeltaYx2MinusDeltaXx2;
        }
        else
        {
            /* add to the error time */
            ErrorTerm += DeltaYx2;
        }

        x0 += xdirection; /* advance the x coordinate */
        drawdot(x0, y0);
    }
}



/****************************************************************/
/* Name: drawdot                                                */
/*                                                              */
/* Description:                                                 */
/*                                                              */
/* draws a dot at x0 and y0                                     */
/*                                                              */
/****************************************************************/
static void drawdot(U32 x0, U32 y0)
{
    /* setup pixel colour */
    U32 colorSDL = SDL_MapRGB(pScr->format, Colour.r, Colour.g, Colour.b);
    /* put screen buffer pointer at the required x & y coord's*/
    U32 *bufp = (U32*)pScr->pixels + y0 * pScr->pitch / 4 + x0;
    /* draw a pixel of colour colorSDL at coord's x & y */
    *bufp = colorSDL;
}
